/*+		T D T 

 * Module name:
     Tdt

 * Function:
     Definitions of items from the tdt namespace.

 * Description:
     This module provides the implementation of items in the tdt namespace,
     as declared in tdt.hh.

 * Language:
      C++

 * Support: {author}, AAO

 *-

 * "@(#) $Id$"

 * History:
     {dd-mmm-yyyy} - {initials} - Original version

 *  Copyright (c) Macquarie University, {year}.
    Not to be used for commercial purposes without permission.
    No warranty of any form unless subject to a specific agreement.

 */
/*
 * RCS id
 */
static const char *rcsId="@(#) $Id$";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include "tdt.hh"


/*
 * Include message definition files from sub=systems of interest.  We
 * should include these for any DRAMA library we use, but DITS, SDS etc.
 * are already included.  These is a corresponding call to MessPutFacility()
 * in the constructor below.  
 */
#include "Git_Err_msgt.h"
#include "drama2_err_msgt.h"
#include "drtf_err_msgt.h"
/* Our own message definition file */

#include "tdt_err_msgt.h"

/*
 * The above are error code definitions, these are the error codes from
 * our sub-system, generated from tdf_err.msg.
 */
#include "tdt_err.h"


/* TASK_DESCR is output by the "tasks" command.  Use it to give a quick 
 * description of your task
 */
static const char *TASK_DESCR  = "TDT Task";

/*
 * GIT Task type string.  Possible values are
 *     CT   Control task
 *     DCDT Detector Control Task 
 *     DHDT Data Handling Task 
 *     DRT Data Recording Task
 *     TEL Telescope control task 
 *     IDT Instrument Task
 *
 * See http://drama.aao.gov.au/doc/ps/git_spec_9.ps for details.
 *
 *  (The "D" above comes from "D-Task", from an older system we used.)
 */
static const std::string TASK_TYPE   = "IDT";
/*
 * Task global buffer size.
 */
static const int GLOBAL_BUF_SIZE = 70000;
/*
 * DRAMA Task flags.  See DitsAppInit() for details of what you might
 * want here.
 */
static const int TASK_FLAGS = 0;

/** Construct the task.
 *
 * @param name The DRAMA name for the task.
 * @param logSysName The log system name.  Passed to drama::Logger 
 *         constructor.
 * @param descr The description of the task for the GIT ENQ_DEV_DESCR
 *         parameter.
 */
tdt::Task::Task(
    const std::string &name,
    const std::string &logSysName,
    const std::string &descr) :
    drama::git::Task(name,                 // DRAMA Task name
                     logSysName,           // Logging system name
                     TdtVersion,           // Task version string (from GIT_REPO_RELEASE macro in Makefile.)
                     TdtDate,              // Task build date (from compiler)
                     descr,                // Task description (ENQ_TASK_DESCR parameter)
                     TASK_TYPE,            // Task type
                     "",                   // Simulation environment variables.
                     GIT_M_SIM_ALLLEVELS,  // Supported simulation levels.
                     GLOBAL_BUF_SIZE,      // Task global buffer size.
                     TASK_FLAGS),          // DRAMA Task flags

    _parSys(this->TaskPtr()),
    _fitsInterface(this),
    _tdtStringParam(TaskPtr(),"STRING_PARAM", "A String"),
    _tdtIntParam(TaskPtr(), "INT_PARAM", 23),
    _tdtShortParam(TaskPtr(), "SHORT_PARAM", 0),
    _tdtDoubleParam(TaskPtr(), "DOUBLE_PARAM", 0),
    _tdtStructParam(TaskPtr(), "STRUCT_PARAM", CreateMyStructParam())
{


    /*
     * Add required message facilities not known by DRAMA DITS level,  We
     * need to have included the right file above.  These calls ensure
     * we can translate error codes to strings.
     */
    MessPutFacility(&MessFac_GIT);
    MessPutFacility(&MessFac_DRAMA2);
    MessPutFacility(&MessFac_DRTF);

    MessPutFacility(&MessFac_TDT);


   /*
     * Add our actions.  Rather then objects for each action (the original
     * DRAMA2 approach), we are using methods of this class, which makes
     * it easy to access our internal information, but does remove
     * the ability of the action to maintain state, other then in the
     * task itself.  (You do need DRAMA 2 version 1.90 or later for
     * this to work - please upgrade if these methods don't exist)
     *
     * Note - in the modern style, INITIALISE and RESET have the
     * same implementation.
     *
     * You can add what ever actions you require and a task may have many
     * actions. These four are required by GIT and provide the example
     * implementations.
     */
    AddMthThd("INITIALISE", &tdt::Task::InitialiseAction, "Initialise task");
    AddMthThd("RESET",      &tdt::Task::InitialiseAction, "Reset task");
    AddMthThd("EXIT",       &tdt::Task::ExitAction,       "Shutdown the task");
    AddMth("REPORT",        &tdt::Task::ReportAction,     "Report on task status");



    /*
     * Set the DRAMA task description for the "tasks" command.
     */
    SetDetails(TASK_DESCR);

}

/*
 * Static Method used to construct the structured parameter.  If
 * you have structured parameters  (template type drama::sds::Id)
 * then you need one of these for each parameter.
 */
drama::sds::IdPtr tdt::Task::CreateMyStructParam()
{
    drama::sds::IdPtr param(
        new drama::sds::Id(drama::sds::Id::CreateTopLevel("STRUCT_PARAM", SDS_STRUCT)));
    /*
     * Example here taken from telescope position structured parameter.
     *  since it uses a lot of types.
     *
     * Anything that can be constructed as a SDS item can be used.
     */
    param->Put("RA",        0.0);    // Apparent RA, degrees
    param->Put("DEC",       0.0);    // Apparent Dec, degrees
    param->Put("AZIMUTH",   0.0);    // Azimuth, degrees.
    param->Put("ELEVATION", 0.0);    // Elevation, degrees.
    param->Put("HA",        0.0);    // HA, degrees
    param->Put("ZD",        0.0);    // ZD, degrees
    param->Put("AIRMASS",   0.0);    // Air mass, atmospheres.
    param->Put("VIGNETTED", (unsigned short)(1));      // Is telescope vignetted.
    param->Put("RA_STR",    "");     // Apparent RA, HH:MM:SS.ss
    param->Put("HA_STR",    "");     // Apparent HA, HH:MM:SS.ss
    param->Put("DEC_STR",   "");     // Apparent Dec, DD:MM:SS.ss
    param->Put("TEL_STATE", "INITIAL"); // Copy of TEL_STATE parameter.
    return param; 
}
 


/*
 * Implement the task's EXIT action.
 *
 * Note that this version does not use the action argument, but it is not
 * uncommon for tasks to have an optional argument to EXIT that might for
 * example, allow a quick (unclean) shutdown rather then normal tidy
 * shutdown.
 */
void tdt::Task::ExitAction(
    drama::thread::TAction *threadAct,
    const drama::sds::Id & /*obeyArg*/)
{
    if (_Initialised)
    {
        _Initialised = false; // Makes it clear to the outside we are not Initialise
    }
    threadAct->MessageUser("Task shutting down");
    /*
     *  This is how we tell the DRAMA task to shutdown.
     */
    threadAct->SetExitOnCompletion();


}

/*
 * This bit is about creating the ResetArgType type below.  It could be
 * in an include file but is not used anywhere else.  We wrap it in the
 * namespace to avoid conflict of the ResetArgLookup class with any other
 * class of the same name.
 */
namespace tdt {
    /**
     * This enum is used to interpret the argument to a RESET action.
     *
     */
    enum ResetArgEnum { Soft=0, Full, Hard, Invalid } ;
    /**
     *  This lookup class is used with ResetArgNum as part of
     * creating a drama::gitarg::Enum item.
     */
    class ResetArgLookup :  drama::gitarg::EnumLookupClass {
    public:
        /** Return the maximum enum value as an integer */
        unsigned int GetMaxValue() const override final {
            return (int)(Invalid)-1; /* We don't accept invalid, want the next one down */
        }
        /** Return the a pointer to the set of string equivalents of each enum */
        const char  ** GetStringArray() const override final {
            static const char *table[] = { "SOFT", "FULL", "HARD", 0 };
            return table;
        }
    };
    /** ResetArgType is a type which can be used to represent a reset
     * action argument.
     *
     */
    using ResetArgType = drama::gitarg::Enum<ResetArgLookup, ResetArgEnum>;
} // namespace tdt



/* Method Implement the task's INITIALISE action as a threaded action.
 */
void tdt::Task::InitialiseAction(
    drama::thread::TAction *threadAct,
    const drama::sds::Id &obeyArg)
{
    /*
     * If we have an argument, then it should indicate the reset type.
     * All we are really interested in is if it is a hard reset.
     */
    bool resetHard = false;
    if (obeyArg)
    {
        tdt::ResetArgType arg(threadAct, obeyArg);
        resetHard =  (arg == tdt::Hard);

        // All other rests are the same type.  But you could do different
        // things with soft/full resets from hard.
    }


    bool DoShutdown = false;
    /*
     * Lock whilst checking state and dealing with being busy.
     */
    {
        drama::thread::AccessDrama dramaLock(*threadAct);
        DoShutdown = _Initialised;
    }
    try
    {
        /*
         * If we have already been initialised, then shutdown.
         */
        if (DoShutdown)
        {
            /*
             * do what ever we need to shutdown!
             */
            ; // Shutdown code common to simulation and non-simulation

            if (_isSimulating)
            {
                ; // Shutdown for simulation case
            }
            else
            {
                ; // Shutdown for non simulation case.
            }
            
            /*
             * Indicate we are no longer initialised.
             */
            {
                drama::thread::AccessDrama dramaLock(*threadAct);
                _Initialised = false;
            }

            Logger().Log(D2LOG_INST, true, "tdt:Task:Init",
                         "TDT Task being reinitialised");

        }
        Logger().Open(GetLogSysName()); // drama::Task and git::Task methods.


        _isSimulating = IsSimulating();   // From git::Task.


        Logger().Log(D2LOG_INST, true, "tdt:Task:Init",
                     (DoShutdown ? "Resetting task." : "Initialising task."));


        /*
         * Do what we need to initialise the task.
         */
        {
            /* Anything which requires a lock can be done after this. */
            drama::thread::AccessDrama dramaLock(*threadAct);
        }
        /*
         * Should be running at this point.
         */
        
        if (_isSimulating)
        {
            threadAct->MessageUser("Have initialised task in simulation");
        }
        else
        {
            threadAct->MessageUser("Have initialised task.");
        }

        /*
         * Indicate we are initialised.  Note - this sets the INITIALISED
         * parameter to true so othertasks know we are done.
         */
        {
            drama::thread::AccessDrama dramaLock(*threadAct);
            _Initialised = true;
        }
        

    }
    catch (...)
    {
        /*
         * If we had an exception, we must insure the flag if false.
         */
        drama::thread::AccessDrama dramaLock(*threadAct);
        _Initialised = false;
        throw;
    }




}

/*
 * Check if the task is initialised, throw a DRAMA exception, with
 * status TDF__NOTINIT if we are no initialised
 */
void tdt::Task::CheckInitialised(const std::string &action)
{
    if (_Initialised) return;

    DramaTHROW_S(TDT__NOTINIT, "Task not yet initialsied when attepting to execute % action",
                 action);
}
 

/*
 * Implement the task's REPORT action - used to allow a user to see
 * some details about the state of the task. Use report using MessageUser().
 */
drama::Request tdt::Task::ReportAction(
    drama::MessageHandler *mh)
{
    if (_isSimulating)
        mh->MessageUser("Task running in simulation mode");
    else
        mh->MessageUser("Task running in real mode");

    return drama::RequestCode::End;

}

