#ifndef _TDT_INC
#define _TDT_INC
/**
 * @file
 *
 * TDT include file.  Comments in this file should be made presuming
 * processing by the Doxygen tool.
 *
 * @author {author}, AAO
 *
 *
 * @copyright (c) Macquarie University, {year}.
 *    Not to be used for commercial purposes without permission.
 *    No warranty of any form unless subject to a specific agreement.
 */
/*
 * History:
    {dd-mmm-yyyy} - {initials} - Original version

 * "@(#) $Id$"
 */

#include "drama.hh"
#include "drama/gittask.hh"
#include "drtf.hh"

namespace tdt {

    class Task; 

    /* These two are defined in the tdfversion.cpp file, which gets the
     * right values at compile time
     */
    /** tdt version number, from GIT_REPO_RELEASE macro in Makefile */
    extern const char *  TdtVersion;
    /** tdt build date, from compiler */
    extern const char *  TdtDate;

    /*
     *  FITS Interface class.  You implement this class to support
     * Adding FITS header keywords from your task.  You can drop it
     * and the include of drtf.h above if you don't have them (You can
     * also drop  IDir(DRTF_DIR) and LinkLibDir(DRTF_LIB,drtf) from your
     * dmakefile.
     *
     * We must implement the drtf::FitsInterface interface, providing the
     * Setup(), Start() and Finish() methods - see the tdffits.cpp file for
     * implementations.
     */
    class TdtFitsClass : public drtf::FitsInterface {
    private:

        drama::ParSys _parSys;  // We need access to the DRAMA parameter system.
        Task *_theTask;   // Access to the task.

        /*
         * Override these methods to implement the required
         * functionality.
         */
        void Setup() override;
        void Start(long int run_id) override;
        void Finish(long int run_id) override;
    public:
        /*
         * Constructor - just invoke the FitsInterface()
         * constructor and grab a reference to the parameter
         * system.
         */
        TdtFitsClass(Task *theTask);

    };

    /** Create a TDT task, which is a sub-class of git::Task.  See the git::Task
     * documentation for more details on GIT tasks. 
     *
     * Also see the GIT documentation - 
     *    http://drama.aao.gov.au/doc/ps/git_spec_9.ps
     *
     * This is used if you are creating a new task
     * type.  Can be removed if this is simply a library. You could also
     * sub-class drama::Task rather then drama::Git::Task or some other
     * Task if required.  But as a GIT task, you get default implementations
     * various actions and parameters.
     */
    class Task : public drama::git::Task {
    private:

        /* Access to the parameter system */
        drama::ParSys _parSys;

        /* FITS header interface */
        TdtFitsClass _fitsInterface;
        /* Are we in simulation mode.? */
        bool _isSimulating = false;

        /*
         * Examples of DRAMA parameters for this task.  
         *   Task authors should replace by your own set
         *
         * Or you can create/access parameters using the _parSys object,
         * above.
         */
        drama::Parameter<std::string>    _tdtStringParam;
        drama::Parameter<INT32>          _tdtIntParam;
        drama::Parameter<short>          _tdtShortParam;
        drama::Parameter<double>         _tdtDoubleParam;
        drama::Parameter<drama::sds::Id> _tdtStructParam;


        /*
         * Check if the task is initialised, throw a DRAMA exception, with
         * status TDF__NOTINIT if we are no initialised
         *
         * @param action Name of action we are trying to execute, will
         *               be added to the exception.
         */
        void CheckInitialised(const std::string &action);
        /*
         * Method used to construct the structured parameter.  If
         * you have structured parameters above (template type drama::sds::Id)
         * then you need one of these for each parameter.
         */
        static drama::sds::IdPtr CreateMyStructParam();
 

        /* Method Implement the task's INITIALISE and RESET actions as a 
         * threaded action.
         */
        void InitialiseAction(drama::thread::TAction *threadAct,
                              const drama::sds::Id &obeyArg);
        /*
         * Method to Implement the task's EXIT action as a threaded action.
         */
        void ExitAction(drama::thread::TAction *threadAct,
                        const drama::sds::Id &obeyArg);

       /*
         * Method to Implement the task's REPORT action as a non-threaded action.
         */
        drama::Request ReportAction(drama::MessageHandler *);

        /*
         * Add your own action implementations and anything else you need.
         */

    public:
        /** Construct the task.
         *
         * @param name The DRAMA name for the task.
         * @param logSysName The log system name.  Passed to drama::Logger 
         *         constructor.
         * @param descr The description of the task for the GIT ENQ_DEV_DESCR
         *         parameter.
         */
        Task(const std::string &name,
             const std::string &logSysName,
             const std::string &descr);
    }; // class Task


} // namespace tdt


#endif // ifndef _TDT_INC
